function [V, S, max_res] = rat_Krylov_type_GenLyap_tangential( A, N, B, m, smin, smax, update_n_step )
%RAT_KRYLOV_TYPE_GENLYAP_TANGENTIAL
%  Using a naive block-Arnoldi to compute a rational Krylov type subspace
%  aimed at approximating the solution to the generalized Lyapunov equation
%             AX + XA^T + \sum_{i=1}^m N{i}XN{i}^T + BB^T = 0.
%  
%  The space created is
%      V = {B, (A-S(1))^{-1}r_1d_1, (A-S(2))^{-1}r_2d_2, ... , (A-S(k))^{-1}r_kd_k},
%  where r_i is a few (typically 10) of the dominating left signular vectors
%  of the residual in the previous space, and d_k is a tangential weight.
%
%  Tangential shift and direction search are, in each step, in accordance with 
%  V. Druskin, V. Simoncini, and M. Zaslavsky. "Adaptive tangential
%  interpolation in rational Krylov subspaces for MIMO dynamical systems".
%  SIAM J. Matrix Anal. Appl., 2014.
%  Specifically (2.3) and Section 3. See also Algorithm 1.
%
%  Returns a space of dimension size(B,2)+(m-1), that is, it does m-1 steps
%
%  [V, S, max_res] = rat_Krylov_type_GenLyap_tangential( A, N, B, m, smin, smax, update_n_step )
%
%  Input:
%  A - Matrix defining the Lyapunov operator
%  N - Cell array containing the "generalized" terms of the equation
%  B - Low-rank factor of the "right-hand-side"
%  m - The method does m-1 steps
%  smin - Estimate of the smallest eigenvalue of A
%  smax - Estimate of the largest eigenvalue of A
%  update_n_step - Allows to update the residual direction r only every
%    update_n_step of the outer iteration. Can be set to 1 if it ise
%    desired to update the residual direction at every iteration
%
%  Output:
%  V - A matrix containing an orthogonal basis of the suggested search
%    space for approximating the generalized Lyapunov equation.
%  s - The used shifts
%  max_res - The maximum residuals at where the corresponding shifts were found
%

if ~iscell(N); error('N needs to be a cell-array'); end

if smax < smin
  error('Input error: smax < smin')
end
if (abs(smin) < 1e-12 && smax == 0) || (abs(smax-smin)/smax < 1e-12)
  s_range = smax;
else
  s_range = linspace(smin, smax, 30);
end

n = size(A,1);
mm = m-1;
r = [];
R = [];

% V = zeros(n, r*m);
S = NaN((mm), 1);
max_res = -inf*ones((mm), 1);

if issparse(A)
  I = speye(n,n);
else
  I = eye(n,n);
end


ta = 10;

V = qr_orth(B); %Orthogonalize and normalize within starting block

k = 1;
while k <= (mm)
  fprintf('Iteration: %2.i of %i\n',k,(mm))
  
  [r, R] = update_residual_direction_GenLyap(A, N, B, V, k, update_n_step, r, R, ta);
  
  % Find a new shift
  % Search the grid for a good point. New point where residual is maximum.
  nn = size(V,2);
  II = speye(nn, nn);
  s = NaN;
  Ap = V'*A*V;
  Rp = V'*r;
  for ss = s_range
    %       res = norm(R - (ss*I - A) * (V * ((ss*II - Ap)\Rp)));
    % Below is an optimized way. Make sure to work with vectors!
    r1 = (V * ((ss*II - Ap)\Rp));
    r3 = A*r1;
    r4 = ss*r1;
    rr = r-r4+r3;
    res = svds(rr,1);%norm(RR);
    if(res > max_res(k))
      max_res(k) = res;
      s = ss;
    end
  end
  %Compute the tangential direction
  r1 = (V * ((s*II - Ap)\Rp));
  r3 = A*r1;
  r4 = s*r1;
  rr = r-r4+r3;

  [~,Sd,d] = svd(rr,0);
  available_cols = mm-k+1;
  Sd = min( [sum(diag(Sd)/Sd(1,1)>0.01), available_cols, ta]);
  d = d(:,1:Sd);
  
  
  idx = (k):(k+Sd-1);
  S(idx) = s;
  max_res(idx) =  max_res(k);

  % Expand the basis
  Vi = (A - s*I)\(r*d);
  V = qr_orth([V,Vi]);
  
  
  % Update index with size of expansion
  k = k + Sd;
end

end