function [r, R] = update_residual_direction_GenLyap(A, N, B, V, k, update_n_step, r, R, ta)
% UPDATE_RESIDUAL_DIRECTION  -  Updates the residual
%  Aims at computing an inner solution to the generalized Lyapunov equation,
%  AX + XA^T + \sum_{i=1}^m N{i}XN{i}^T + BB^T = 0,
%  projected on the subspace spanned by V (assumed to be Orthogonal).
%  This is achieved by a call to INNER_SOLVER_GENLYAP.
%  Based on this it computes the dominating left-singular vector r of the
%  residual R = AZ + ZA^T + \sum_{i=1}^m N{i}ZN{i}^T + BB^T,
%  where Z is the computed approximation Z = V*Y*V'.
%
%  OBS: Allowed to be optimized depending on problem structure!
%
%  Targets generalized Lyapunov equations: AX + XA^T + NXN^T + BB^T = 0,
%  but not restricted to that. Finds shifts for spaces of the type
%  {B, (A-S(1))^{-1}w_1, (A-S(2))^{-1}w_2, ... , (A-S(k))^{-1}w_k}
%
%  [r, R] = update_residual_direction_GenLyap(A, N, B, V, k, update_n_step, r, R)
%
%  Input:
%  A - Matrix defining the Lyapunov operator
%  N - Cell array containing the "generalized" terms of the equation
%  B - Low-rank factor of the "right-hand-side"
%  V - Orthogonal matrix containing a basis of the subpsace projected on
%  k - Current step in the outer iteration. Can be set to 0 if not desired
%  update_n_step - Allows to update only every update_n_step. Regulated
%    together with k above. Can be set to 1 if not desired
%  r - The previous r. Not needed if k == 0 and/or update_n_step = 1
%  R - The previous R. Not needed if k == 0 and/or update_n_step = 1
%  ta - Adaptation for tangential approaches. Computes the ta dominant left
%    singular vectors of the residual. Default is 1. Can be set to 0 if r
%    is not needed to save some computation effort.
%
%  Output:
%  r - The left singular vector corresponding to the largest singular value
%    of the full residual R. Or the ta ones, see the input.
%  R - The full residual R
% 

  if ~iscell(N); error('N needs to be a cell-array'); end;
  if (mod(k,update_n_step) == 0) || (k == 1)
    % Exactly by solving the inner problem
    Y = inner_solver_GenLyap( A, N, B, V );

    AA = A*V;
    R = (((AA)*Y)*V') + (V*(Y*(AA'))) + B*B';
    for i = 1:length(N)
      NN_i = N{i}*V;
      R = R + (NN_i)*Y*(NN_i');
    end
    
    if ~exist('ta', 'var') || isempty(ta) || (ta == 1)
      [VV,~,~] = svds(R,1);
      r = VV(:,1);
    elseif (ta == 0)
      r = [];
    else
      [VV,SS,~,~] = svds(R,ta);
      nn = min(ta,size(SS,2));
      r = VV(:,1:nn)*SS;%*WW(:,1:nn)';
    end
    
    
% %     % JUST A FASTER WAY TO COMPUTE THE SVD ABOVE
% % %     W = [AA, V, NN, B];
% % %     WW = [V, AA, NN, B];
% % %     [Q_W,R_W] = qr(W,0);
% % %     [~ ,R_WW] = qr(WW,0);
% % %     T = R_W * blkdiag(Y,Y,Y,eye(size(B,2))) * R_WW';
% % %     [UT,~,~] = svd(T);
% % %     R = Q_W*UT(:,1);
  else
    % Do not update the residual in each step, to save computation time
  end


end